export const Self = Symbol()
export const Value = Symbol()

export function constant() { return this }

/**
 * @template T, U
 * @typedef {(x: T) => U} Morphism
 */

/**
 * @template T
 * @typedef {{ [Value]: T }} Set
 */

/**
 * @template T
 * @this {Construct<T>}
 * @typedef {{ [Self]: <a>(value: a) => Construct<a> }} Construct
 */

/**
 * @template T
 * @param {T} value
 * @returns {Set<T>}
 */
export const Set = value => ({ [Value]: value })

/**
 * @template T, U
 * @param {Morphism<T, U>} fn 
 * @returns {U}
 */
export function chain(fn) {
  return fn(this[Value])
}

/**
 * @template T
 * @this {Set<T>}
 * @returns {T}
 */
export function extract() {
  return this[Value]
}

/**
 * @template T, U
 * @this {Construct<T>}
 * @param {U} value 
 * @returns {Construct<U>}
 */
export function of(value) {
  return this[Self](value)
}

